import unittest
import pandas as pd
import numpy as np
from numpy.testing import assert_array_almost_equal
from pandas.testing import assert_frame_equal
from nancorrmp.nancorrmp import NaNCorrMp
from scipy.stats import pearsonr


class TestNaNCorrMp(unittest.TestCase):
    X = pd.DataFrame({'a': [1, 5, 7, 9, 4], 'b': [-1, 3, -3, 1, 4], 'c': [0, 1, -1, -2, 4], 'd': [-5, -3, -2, 1, 4]})
    X_nans = pd.DataFrame({'a': [float('NaN'), 5, 7, 9], 'b': [-1, 3, float('NaN'), 1], 'c': [0, 1, -1, -2], 'd': [-5, -3, -2, 1]})
    X_infs = pd.DataFrame({'a': [float('-inf'), 5, 7, 9], 'b': [-1, 3, -3, 1], 'c': [0, 1, -1, -2], 'd': [-5, -3, -2, float(('+inf'))]})

    def test_without_nans(self) -> None:
        result = NaNCorrMp.calculate(self.X, n_jobs=2, chunks=1)
        expected_result = self.X.corr()
        assert_frame_equal(result, expected_result)

    def test_with_nans(self) -> None:
        result = NaNCorrMp.calculate(self.X_nans, n_jobs=2, chunks=1)
        expected_result = self.X_nans.corr()
        assert_frame_equal(result, expected_result)

    def test_with_numpy_input(self) -> None:
        result = NaNCorrMp.calculate(self.X_nans.to_numpy().transpose(), n_jobs=2, chunks=1)
        self.assertEqual(type(result), np.ndarray)
        expected_result = self.X_nans.corr()
        assert_array_almost_equal(result, expected_result.to_numpy().transpose())

    def test_with_infs(self) -> None:
        result = NaNCorrMp.calculate(self.X_infs, n_jobs=2, chunks=1)
        expected_result = self.X_infs.corr()
        assert_frame_equal(result, expected_result)

    def test_single_core_without_nans(self) -> None:
        result = NaNCorrMp.calculate(self.X, n_jobs=1)
        expected_result = self.X.corr()
        assert_frame_equal(result, expected_result)

    def test_single_core_with_nans(self) -> None:
        result = NaNCorrMp.calculate(self.X_nans, n_jobs=1)
        expected_result = self.X_nans.corr()
        assert_frame_equal(result, expected_result)

    def test_single_core_with_infs(self) -> None:
        result = NaNCorrMp.calculate(self.X_infs, n_jobs=1, chunks=1)
        expected_result = self.X_infs.corr()
        assert_frame_equal(result, expected_result)

    def test_calculate_with_p_value_without_nans(self) -> None:
        correlations, p_values = NaNCorrMp.calculate_with_p_value(self.X, n_jobs=2, chunks=1)
        empty_dataframe = pd.DataFrame(columns=correlations.columns, index=correlations.index, copy=True, dtype=np.float64)
        expected_correlations, expected_p_values = empty_dataframe.copy(), empty_dataframe.copy()
        for column in self.X.columns:
            for other_column in self.X.columns:
                expected_correlation, expected_p_value = pearsonr(self.X[column], self.X[other_column])
                expected_correlations[column][other_column] = expected_correlation
                expected_p_values[column][other_column] = expected_p_value
        assert_frame_equal(correlations, expected_correlations)
        assert_frame_equal(p_values, expected_p_values)

    def test_calculate_with_p_value_with_nans(self) -> None:
        correlations, p_values = NaNCorrMp.calculate_with_p_value(self.X_nans, n_jobs=2, chunks=1)
        self.assertFalse(correlations.isnull().values.any())
        self.assertFalse(p_values.isnull().values.any())



import unittest
import numpy as np
import pandas as pd
import time
from numpy.testing import assert_array_almost_equal
from numpy.random import randint

from nancorrmp.nancorrmp import NaNCorrMp


class TestBenchmarkNaNCorrMp(unittest.TestCase):

    @unittest.skip("computation to long")
    def test_benchmark_with_pandas_without_nans(self) -> None:
        # faster when size > 1200, n_jobs=4
        size = 1200
        n_jobs = 4
        np.random.seed(0)
        random_dataframe = pd.DataFrame(np.random.rand(size, size))
        t_start = time.perf_counter()
        nancorrmp_result = NaNCorrMp.calculate(random_dataframe, n_jobs=n_jobs, chunks=1000)
        nancorrmp_time = time.perf_counter() - t_start
        print(f'NaNCorrMp time with {n_jobs} jobs: {nancorrmp_time}')
        t_start = time.perf_counter()
        pandas_result = random_dataframe.corr()
        pandas_time = time.perf_counter() - t_start
        print(f'pandas time: {pandas_time}')
        print(f'nancorrmp_time / pandas_time ratio: {nancorrmp_time/pandas_time}')
        assert_array_almost_equal(nancorrmp_result.to_numpy(), pandas_result.to_numpy())
        self.assertTrue(nancorrmp_time < pandas_time)

    @unittest.skip("computation to long")
    def test_benchmark_with_pandas_with_nans(self) -> None:
        # faster when size > 1200, n_jobs=4
        size = 5000
        n_jobs = 4
        np.random.seed(0)
        nan_infs_ratio = 0.03
        random_dataframe = self._get_random_dataframe_with_nans_and_infs(size, nan_infs_ratio)
        t_start = time.perf_counter()
        nancorrmp_result = NaNCorrMp.calculate(random_dataframe, n_jobs=n_jobs, chunks=1000)
        nancorrmp_time = time.perf_counter() - t_start
        print(f'NaNCorrMp time with {n_jobs} jobs: {nancorrmp_time}')
        t_start = time.perf_counter()
        pandas_result = random_dataframe.corr()
        pandas_time = time.perf_counter() - t_start
        print(f'pandas time: {pandas_time}')
        print(f'nancorrmp_time / pandas_time ratio: {nancorrmp_time / pandas_time}')
        assert_array_almost_equal(nancorrmp_result.to_numpy(), pandas_result.to_numpy())
        self.assertTrue(nancorrmp_time < pandas_time)

    @staticmethod
    def _get_random_dataframe_with_nans_and_infs(size: int, nan_infs_ratio: float) -> pd.DataFrame:
        possible_values = (float('NaN'), float('+inf'), float('-inf'))
        random_dataframe = pd.DataFrame(np.random.rand(size, size))
        for i in range(int(nan_infs_ratio * size * (size - 1) / 2)):
            value = possible_values[i % 3]
            x, y = randint(size), randint(size)
            random_dataframe[x][y] = value
        return random_dataframe

    @unittest.skip("computation to long")
    def test_benchmark_with_numpy_without_nans(self) -> None:
        size = 1200
        n_jobs = 4
        np.random.seed(0)
        random_dataframe = pd.DataFrame(np.random.rand(size, size))
        t_start = time.perf_counter()
        nancorrmp_result = NaNCorrMp.calculate(random_dataframe, n_jobs=n_jobs, chunks=1000)
        nancorrmp_time = time.perf_counter() - t_start
        print(f'NaNCorrMp time with {n_jobs} jobs: {nancorrmp_time}')
        t_start = time.perf_counter()
        numpy_result = np.corrcoef(random_dataframe.to_numpy().transpose())
        numpy_time = time.perf_counter() - t_start
        print(f'numpy time: {numpy_time}')
        print(f'nancorrmp_time / numpy_time ratio: {nancorrmp_time/numpy_time}')
        assert_array_almost_equal(nancorrmp_result.to_numpy(), numpy_result)
        self.assertTrue(nancorrmp_time < numpy_time)



import ctypes
from multiprocessing.sharedctypes import RawArray
import numpy as np
import pandas as pd
import multiprocessing as mp
from typing import *
from scipy.special import btdtr

shared_variables_dictionary = {}

ArrayLike = Union[pd.DataFrame, np.ndarray]


class NaNCorrMp:

    @staticmethod
    def _init_worker(X: RawArray, X_finite_mask: RawArray, X_corr: RawArray,
                     X_shape: Tuple[int, int], X_corr_shape: Tuple[int, int],
                     X_p_value: RawArray = None) -> None:
        shared_variables_dictionary['X'] = X
        shared_variables_dictionary['X_finite_mask'] = X_finite_mask
        shared_variables_dictionary['X_corr'] = X_corr
        shared_variables_dictionary['X_shape'] = X_shape
        shared_variables_dictionary['X_corr_shape'] = X_corr_shape
        shared_variables_dictionary['X_p_value'] = X_p_value
        shared_variables_dictionary['X_p_value_shape'] = X_corr_shape

    @staticmethod
    def calculate(X: ArrayLike, n_jobs: int = -1, chunks: int = 500) -> ArrayLike:
        return NaNCorrMp._calculate(X=X, n_jobs=n_jobs, chunks=chunks, add_p_values=False)

    @staticmethod
    def calculate_with_p_value(X: ArrayLike, n_jobs: int = -1, chunks: int = 500) -> Tuple[ArrayLike, ArrayLike]:
        return NaNCorrMp._calculate(X=X, n_jobs=n_jobs, chunks=chunks, add_p_values=True)

    @staticmethod
    def _calculate(X: ArrayLike, n_jobs: int, chunks: int, add_p_values: int) -> Union[ArrayLike, Tuple[ArrayLike, ArrayLike]]:
        X_array = X.to_numpy(dtype=np.float64, copy=True).transpose() if type(X) == pd.DataFrame else X
        X_raw = RawArray(ctypes.c_double, X_array.shape[0] * X_array.shape[1])
        X_np = np.frombuffer(X_raw, dtype=np.float64).reshape(X_array.shape)
        np.copyto(X_np, X_array)

        finite_mask_data = np.isfinite(X_array)
        finite_mask_raw = RawArray(ctypes.c_bool, X_array.shape[0] * X_array.shape[1])
        finite_mask_np = np.frombuffer(finite_mask_raw, dtype=bool).reshape(X_array.shape)
        np.copyto(finite_mask_np, finite_mask_data)

        X_corr = np.ndarray(shape=(X_array.shape[0], X_array.shape[0]), dtype=np.float64)
        X_corr_raw = RawArray(ctypes.c_double, X_corr.shape[0] * X_corr.shape[1])
        X_corr_np = np.frombuffer(X_corr_raw, dtype=np.float64).reshape(X_corr.shape)

        if add_p_values:
            X_p_value = np.ndarray(shape=X_corr.shape, dtype=np.float64)
            X_p_value_raw = RawArray(ctypes.c_double, X_p_value.shape[0] * X_p_value.shape[1])
            X_p_value_np = np.frombuffer(X_p_value_raw, dtype=np.float64).reshape(X_corr.shape)
        else:
            X_p_value_np = None
            X_p_value_raw = None
            X_p_value_np = None

        arguments = ((j, i) for i in range(X_array.shape[0]) for j in range(i))
        processes = n_jobs if n_jobs > 0 else None
        worker_function = NaNCorrMp._set_correlation_with_p_value if add_p_values else NaNCorrMp._set_correlation
        with mp.Pool(processes=processes,
                     initializer=NaNCorrMp._init_worker,
                     initargs=(X_raw, finite_mask_raw, X_corr_raw, X_np.shape, X_corr_np.shape, X_p_value_raw)) \
                as pool:
            list(pool.imap_unordered(worker_function, arguments, chunks))

        for i in range(X_corr_np.shape[0]):
            X_corr_np[i][i] = 1.0

        if add_p_values:
            if type(X) == pd.DataFrame:
                return (
                    pd.DataFrame(X_corr_np, columns=X.columns, index=X.columns),
                    pd.DataFrame(X_p_value_np, columns=X.columns, index=X.columns)
                )
            else:
                return X_corr_np, X_p_value_np

        if type(X) == pd.DataFrame:
            return pd.DataFrame(X_corr_np, columns=X.columns, index=X.columns)
        else:
            return X_corr_np

    @staticmethod
    def _set_correlation(arguments: Tuple[int, int]) -> None:
        j, i = arguments
        X_np, X_corr_np, finite_mask = NaNCorrMp._get_global_variables()
        finites = finite_mask[i] & finite_mask[j]
        x = X_np[i][finites]
        y = X_np[j][finites]
        corr = NaNCorrMp._corr(x, y)
        X_corr_np[i][j] = corr
        X_corr_np[j][i] = corr

    @staticmethod
    def _get_global_variables(get_p_value: bool = False) -> Union[Tuple[np.ndarray, np.ndarray, np.ndarray],
                                                                  Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]]:
        X_np = np.frombuffer(shared_variables_dictionary['X'], dtype=np.float64).reshape(shared_variables_dictionary['X_shape'])
        X_corr_np = np.frombuffer(shared_variables_dictionary['X_corr'], dtype=np.float64).reshape(shared_variables_dictionary['X_corr_shape'])
        finite_mask = np.frombuffer(shared_variables_dictionary['X_finite_mask'], dtype=bool).reshape(shared_variables_dictionary['X_shape'])
        if not get_p_value:
            return X_np, X_corr_np, finite_mask
        else:
            X_p_value_np = np.frombuffer(shared_variables_dictionary['X_p_value'], dtype=np.float64).reshape(shared_variables_dictionary['X_p_value_shape'])
            return X_np, X_corr_np, finite_mask, X_p_value_np

    @staticmethod
    def _corr(x: np.ndarray, y: np.ndarray) -> float:
        mx, my = x.mean(), y.mean()
        xm, ym = x - mx, y - my
        r_num = np.add.reduce(xm * ym)
        r_den = np.sqrt((xm * xm).sum() * (ym * ym).sum())
        r = r_num / r_den
        return max(min(r, 1.0), -1.0)

    @staticmethod
    def _set_correlation_with_p_value(arguments: Tuple[int, int]) -> None:
        j, i = arguments
        X_np, X_corr_np, finite_mask, X_p_value_np = NaNCorrMp._get_global_variables(get_p_value=True)
        finites = finite_mask[i] & finite_mask[j]
        x = X_np[i][finites]
        y = X_np[j][finites]
        corr = NaNCorrMp._corr(x, y)
        X_corr_np[i][j] = corr
        X_corr_np[j][i] = corr
        p_value = NaNCorrMp._p_value(corr, len(x))
        X_p_value_np[i][j] = p_value
        X_p_value_np[j][i] = p_value

    @staticmethod
    def _p_value(corr: float, observation_length: int) -> float:
        ab = observation_length / 2 - 1
        if ab == 0:
            p_value = 1.0
        else:
            p_value = 2 * btdtr(ab, ab, 0.5 * (1 - abs(np.float64(corr))))
        return p_value



from setuptools import setup
import sys

if not sys.version_info[0] == 3 and sys.version_info[1] < 6:
    sys.exit('Python < 3.6 is not supported')

version = '0.23'

setup(
    name='nancorrmp',
    packages=['nancorrmp', 'test'],
    version=version,
    description='A multiprocessing version of pandas correlation method',
    author='Michał Bukowski',
    author_email='michal.bukowski@buksoft.pl',
    license='MIT',
    url='https://github.com/bukson/nancorrmp',
    download_url='https://github.com/bukson/nancorrmp/tarball/' + version,
    keywords=['correlation', 'multiprocessing', 'pandas'],
    classifiers=[],
    install_requires=[
        'pandas',
        'numpy',
    ],
    tests_require=[
        'scipy'
    ],
)
